#ifndef ZonedDateTime_D_H
#define ZonedDateTime_D_H

#include <stdio.h>
#include <stdint.h>
#include <stddef.h>
#include <stdbool.h>
#include "diplomat_runtime.h"





typedef struct ZonedDateTime ZonedDateTime;




#endif // ZonedDateTime_D_H
