#ifndef ParsedZonedDateTime_D_H
#define ParsedZonedDateTime_D_H

#include <stdio.h>
#include <stdint.h>
#include <stddef.h>
#include <stdbool.h>
#include "diplomat_runtime.h"





typedef struct ParsedZonedDateTime ParsedZonedDateTime;




#endif // ParsedZonedDateTime_D_H
