// Copyright 2017 The Chromium Authors
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

// Slightly adapted for inclusion in V8.
// Copyright 2025 the V8 project authors. All rights reserved.

#ifndef V8_BASE_NUMERICS_SAFE_MATH_H_
#define V8_BASE_NUMERICS_SAFE_MATH_H_

#include "src/base/numerics/checked_math.h"      // IWYU pragma: export
#include "src/base/numerics/clamped_math.h"      // IWYU pragma: export
#include "src/base/numerics/safe_conversions.h"  // IWYU pragma: export

#endif  // V8_BASE_NUMERICS_SAFE_MATH_H_
