.text	

.globl	_ossl_rsaz_amm52x40_x1_avxifma256
.globl	_ossl_rsaz_amm52x40_x2_avxifma256
.globl	_ossl_extract_multiplier_2x40_win5_avx

_ossl_rsaz_amm52x40_x1_avxifma256:
_ossl_rsaz_amm52x40_x2_avxifma256:
_ossl_extract_multiplier_2x40_win5_avx:
.byte	0x0f,0x0b
	.byte	0xf3,0xc3

