#ifndef PartialTime_H
#define PartialTime_H

#include <stdio.h>
#include <stdint.h>
#include <stddef.h>
#include <stdbool.h>
#include "diplomat_runtime.h"


#include "PartialTime.d.h"










#endif // PartialTime_H
